"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import copy
import numpy as np
import tensorflow as tf
from keras import models
import keras.layers as KL

# third-party imports
from lamareg.ext.lab2im import layers as layers
from lamareg.ext.lab2im import utils, edit_volumes
from lamareg.ext.neuron import models as nrn_models
from lamareg.ext.lab2im import edit_tensors as l2i_et


def sample_segmentation_pairs(
    image_dir,
    labels_dir,
    results_dir,
    n_examples,
    path_model,
    segmentation_labels,
    n_neutral_labels=None,
    batchsize=1,
    flipping=True,
    scaling_bounds=0.15,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_scale=0.04,
    min_res=1.0,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    noise_std_lr=3.0,
    blur_range=1.03,
    bias_field_std=0.5,
    bias_scale=0.025,
    noise_std=10,
    gamma_std=0.5,
):
    """
    This function enables us to obtain segmentations from a segmentation network along with the corresponding ground
    truths. The segmentations are obtained by taking real images and aggressively augmenting them with spatial and
    intensity transforms, such that the network commits some mistakes. The sample pairs can then be used to train a
    denoising network by using the noisy segmentations as inputs and the ground truth as targets.

    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.

    :param image_dir: path of folder with all images
    :param labels_dir: path of folder with all ground truth segmentations
    :param results_dir: path of a directory where the results will be saved
    :param n_examples: number of pairs to sample
    :param path_model: path of the trained model to obtain segmentations from

    # --------------------------------------------- Degradation parameters ---------------------------------------------
    # label maps parameters
    :param n_neutral_labels: (optional) if the label maps contain some right/left specific labels and if flipping is
    applied during training, please provide the number of non-sided labels (including the background).
    This is used to know where the sided labels start in generation_labels. Leave to default (None) if either one of the
    two conditions is not fulfilled.
    :param segmentation_labels: (optional) list of the same length as generation_labels to indicate which values to use
    in the training label maps, i.e. all occurrences of generation_labels[i] in the input label maps will be converted
    to output_labels[i] in the returned label maps. Examples:
    Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
    Set output_labels[i]=generation_labels[i] if you wish to keep the value generation_labels[i] in the returned maps.
    Can be a list or a 1d numpy array, or the path to such an array. Default is output_labels = generation_labels.

    # output-related parameters
    :param batchsize: (optional) number of images to generate per mini-batch. Default is 1.

    # spatial deformation parameters
    :param flipping: (optional) whether to introduce right/left random flipping. Default is True.
    :param scaling_bounds: (optional) if apply_linear_trans is True, the scaling factor for each dimension is
    sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    (1-scaling_bounds, 1+scaling_bounds) for each dimension.
    2) the path to a numpy array of shape (2, n_dims), in which case the scaling factor in dimension i is sampled from
    the uniform distribution of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    3) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for case 1 the
    bounds are centred on 0 rather than 1, i.e. (0+rotation_bounds[i], 0-rotation_bounds[i]).
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
    encourage using it when cropping is deactivated (i.e. when output_shape=None).
    :param nonlin_std: (optional) Standard deviation of the normal distribution from which we sample the first
    tensor for synthesising the deformation field. Set to 0 to completely deactivate elastic deformation.
    :param nonlin_scale: (optional) Ratio between the size of the input label maps and the size of the sampled
    tensor for synthesising the elastic deformation field.

    # blurring/resampling parameters
    :param min_res: (optional) resolution of the input pairs of images/segmentations
    :param max_res_iso: (optional) This enables to control the upper bound of the uniform distribution from which we
    sample the random resolution U(min_res, max_res_iso), where min_res is the resolution of the input label maps.
    Must be a number, and default is 4.
    :param max_res_aniso: This enables to downsample the input volumes to a random LR in only 1 (random) direction. This
    is done by randomly selecting a direction i in the range [0, n_dims-1], and sampling a value in the corresponding
    uniform distribution U(min_res[i], max_res_aniso[i]), where min_res is the resolution of the input label maps. Can
    be a number, a sequence, or a 1d numpy array.
    :param blur_range: (optional) coef to randomise the blurring kernel

    # bias field parameters
    :param bias_field_std: (optional) If strictly positive, this triggers the corruption of images with a bias field.
    The bias field is obtained by sampling a first small tensor from a normal distribution, resizing it to
    full size, and rescaling it to positive values by taking the voxel-wise exponential. bias_field_std designates the
    std dev of the normal distribution from which we sample the first tensor.
    Set to 0 to completely deactivate bias field corruption.
    :param bias_scale: (optional) If bias_field_std is not False, this designates the ratio between the size of
    the input label maps and the size of the first sampled tensor for synthesising the bias field.

    # noise parameters
    :param noise_std: (optional) standard deviation of the white Gaussian noise added at high resolution to the image.
    :param noise_std_lr: (optional) sta dev of the white Gaussian noise added at low resolution to the image.
    :param gamma_std: (optional) standard deviation of the gaussian transform to apply at high resolution to the image.
    """

    # prepare data files
    path_images = utils.list_images_in_folder(image_dir)
    path_labels = utils.list_images_in_folder(labels_dir)
    assert len(path_images) == len(
        path_labels
    ), "There should be as many images as label maps."

    # prepare results subfolders
    gt_result_dir = os.path.join(results_dir, "labels_gt")
    pred_result_dir = os.path.join(results_dir, "labels_seg")
    utils.mkdir(gt_result_dir)
    utils.mkdir(pred_result_dir)

    # get label lists
    segmentation_labels, _ = utils.get_list_labels(
        label_list=segmentation_labels, labels_dir=labels_dir
    )
    n_labels = np.size(np.unique(segmentation_labels))

    # create augmentation model
    im_shape, _, n_dims, n_channels, _, atlas_res = utils.get_volume_info(
        path_images[0], aff_ref=np.eye(4)
    )
    augmentation_model = build_augmentation_model(
        im_shape,
        n_channels,
        segmentation_labels,
        n_neutral_labels,
        n_dims,
        atlas_res,
        flipping=flipping,
        aff=np.eye(4),
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_shape_factor=nonlin_scale,
        min_res=min_res,
        max_res_iso=max_res_iso,
        max_res_aniso=max_res_aniso,
        noise_std_lr=noise_std_lr,
        blur_range=blur_range,
        bias_field_std=bias_field_std,
        bias_shape_factor=bias_scale,
        noise_std=noise_std,
        gamma_std=gamma_std,
    )
    unet_input_shape = augmentation_model.output[0].get_shape().as_list()[1:]

    # prepare the segmentation model
    unet_model = nrn_models.unet(
        nb_features=24,
        input_shape=unet_input_shape,
        nb_levels=5,
        conv_size=3,
        nb_labels=n_labels,
        feat_mult=2,
        nb_conv_per_level=2,
        batch_norm=-1,
        activation="elu",
        input_model=augmentation_model,
    )

    # get generative model
    train_example_gen = build_model_inputs(path_images, path_labels, batchsize)
    segmentation_labels = np.unique(segmentation_labels)

    # redefine model to output deformed image, deformed GT labels, and predicted labels
    list_output_tensors = [unet_model.get_layer("labels_out").output, unet_model.output]
    generation_model = models.Model(
        inputs=unet_model.inputs, outputs=list_output_tensors
    )
    generation_model.load_weights(path_model, by_name=True)

    # generate !
    n = len(str(n_examples))
    i = 1
    while i <= n_examples:

        # predict new segmentation
        outputs = generation_model.predict(next(train_example_gen))

        # save results
        for output, name, res_dir in zip(
            outputs,
            ["labels_gt", "labels_pred_argmax_convert"],
            [gt_result_dir, pred_result_dir],
        ):
            for b in range(batchsize):
                tmp_name = copy.deepcopy(name)
                tmp_output = np.squeeze(output[b, ...])
                if "_argmax" in tmp_name:
                    tmp_output = tmp_output.argmax(-1)
                    tmp_name = tmp_name.replace("_argmax", "")
                if "_convert" in tmp_name:
                    tmp_output = segmentation_labels[tmp_output]
                    tmp_name = tmp_name.replace("_convert", "")
                path = os.path.join(
                    res_dir, tmp_name + "_%.{}d".format(n) % i + ".nii.gz"
                )
                if batchsize > 1:
                    path = path.replace(".nii.gz", "_%s.nii.gz" % (b + 1))
                utils.save_volume(tmp_output, np.eye(4), None, path)

        i += 1


def build_augmentation_model(
    im_shape,
    n_channels,
    segmentation_labels,
    n_neutral_labels,
    n_dims,
    atlas_res,
    flipping=True,
    aff=None,
    scaling_bounds=0.15,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_shape_factor=0.0625,
    min_res=1.0,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    noise_std_lr=3.0,
    blur_range=1.03,
    bias_field_std=0.5,
    bias_shape_factor=0.025,
    noise_std=10,
    gamma_std=0.5,
):

    # define model inputs
    image_input = KL.Input(shape=im_shape + [n_channels], name="image_input")
    labels_input = KL.Input(shape=im_shape + [1], name="labels_input", dtype="int32")

    # deform labels
    labels, image = layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_shape_factor,
        inter_method=["nearest", "linear"],
    )([labels_input, image_input])

    # flipping
    if flipping:
        assert aff is not None, "aff should not be None if flipping is True"
        labels, image = layers.RandomFlip(
            edit_volumes.get_ras_axes(aff, n_dims)[0],
            [True, False],
            segmentation_labels,
            n_neutral_labels,
        )([labels, image])

    # apply bias field
    if bias_field_std > 0:
        image = layers.BiasFieldCorruption(bias_field_std, bias_shape_factor, False)(
            image
        )

    # intensity augmentation
    image = layers.IntensityAugmentation(
        noise_std, gamma_std=gamma_std, contrast_inversion=True
    )(image)

    # loop over channels
    channels = list()
    split = (
        KL.Lambda(lambda x: tf.split(x, [1] * n_channels, axis=-1))(image)
        if (n_channels > 1)
        else [image]
    )
    for i, channel in enumerate(split):

        # reformat resolution range parameters
        min_res = np.array(
            utils.reformat_to_list(min_res, length=n_dims, dtype="float")
        )
        max_res_iso = np.array(
            utils.reformat_to_list(max_res_iso, length=n_dims, dtype="float")
        )
        max_res_aniso = np.array(
            utils.reformat_to_list(max_res_aniso, length=n_dims, dtype="float")
        )
        max_res = np.maximum(max_res_iso, max_res_aniso)

        # sample resolution and thickness (blurring res)
        resolution, blur_res = layers.SampleResolution(
            min_res, max_res_iso, max_res_aniso
        )(channel)
        sigma = l2i_et.blurring_sigma_for_downsampling(
            atlas_res, resolution, thickness=blur_res
        )

        # blur and downsample/resample
        channel = layers.DynamicGaussianBlur(
            0.75 * max_res / np.array(atlas_res), blur_range
        )([channel, sigma])
        channel = layers.MimicAcquisition(
            atlas_res, min_res, im_shape, False, noise_std_lr
        )([channel, resolution])
        channels.append(channel)

    # concatenate all channels back
    image = (
        KL.Lambda(lambda x: tf.concat(x, -1))(channels)
        if len(channels) > 1
        else channels[0]
    )

    # build model (dummy layer enables to keep the labels when plugging this model to other models)
    labels = KL.Lambda(lambda x: tf.cast(x, dtype="int32"), name="labels_out")(labels)
    image = KL.Lambda(lambda x: x[0], name="image_out")([image, labels])
    brain_model = models.Model(
        inputs=[image_input, labels_input], outputs=[image, labels]
    )

    return brain_model


def build_model_inputs(path_images, path_label_maps, batchsize=1):

    # get label info
    _, _, n_dims, n_channels, _, _ = utils.get_volume_info(path_images[0])

    # Generate!
    while True:

        # randomly pick as many images as batchsize
        indices = np.random.randint(len(path_label_maps), size=batchsize)

        # initialise input lists
        list_images = list()
        list_label_maps = list()

        for idx in indices:

            # add image
            image = utils.load_volume(path_images[idx], aff_ref=np.eye(4))
            if n_channels > 1:
                list_images.append(utils.add_axis(image, axis=0))
            else:
                list_images.append(utils.add_axis(image, axis=[0, -1]))

            # add labels
            labels = utils.load_volume(
                path_label_maps[idx], dtype="int", aff_ref=np.eye(4)
            )
            list_label_maps.append(utils.add_axis(labels, axis=[0, -1]))

        # build list of inputs of augmentation model
        list_inputs = [list_images, list_label_maps]
        if batchsize > 1:  # concatenate individual input types if batchsize > 1
            list_inputs = [np.concatenate(item, 0) for item in list_inputs]
        else:
            list_inputs = [item[0] for item in list_inputs]

        yield list_inputs



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
import tensorflow as tf
import keras.layers as KL
from keras.models import Model

# project imports
from lamareg.SynthSeg import evaluate
from lamareg.SynthSeg.predict import write_csv

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.lab2im import edit_volumes
from lamareg.ext.neuron import models as nrn_models


def predict(
    path_images,
    path_masks,
    path_segmentations,
    path_model,
    labels_segmentation,
    labels_mask,
    path_posteriors=None,
    path_volumes=None,
    names_segmentation=None,
    min_pad=None,
    cropping=None,
    sigma_smoothing=0.5,
    strict_masking=False,
    keep_biggest_component=True,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    gt_folder=None,
    evaluation_labels=None,
    list_incorrect_labels=None,
    list_correct_labels=None,
    compute_distances=False,
    recompute=True,
    verbose=True,
):

    # prepare input/output filepaths
    (
        path_images,
        path_masks,
        path_segmentations,
        path_posteriors,
        path_volumes,
        compute,
        unique_vol_file,
    ) = prepare_output_files(
        path_images,
        path_masks,
        path_segmentations,
        path_posteriors,
        path_volumes,
        recompute,
    )

    # get label list
    labels_mask, _ = utils.get_list_labels(label_list=labels_mask)
    mask_labels_unique = np.unique(labels_mask)
    labels_segmentation, _ = utils.get_list_labels(label_list=labels_segmentation)
    labels_segmentation, indices = np.unique(labels_segmentation, return_index=True)

    # prepare volumes if necessary
    if unique_vol_file & (path_volumes[0] is not None):
        write_csv(path_volumes[0], None, True, labels_segmentation, names_segmentation)

    # build network
    _, _, n_dims, n_channels, _, _ = utils.get_volume_info(path_images[0])
    model_input_shape = [None] * n_dims + [n_channels]
    net = build_model(
        path_model=path_model,
        input_shape=model_input_shape,
        labels_segmentation=labels_segmentation,
        n_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        unet_feat_count=unet_feat_count,
        feat_multiplier=feat_multiplier,
        activation=activation,
        sigma_smoothing=sigma_smoothing,
        mask_labels=mask_labels_unique,
    )

    # set cropping/padding
    if (cropping is not None) & (min_pad is not None):
        cropping = utils.reformat_to_list(cropping, length=n_dims, dtype="int")
        min_pad = utils.reformat_to_list(min_pad, length=n_dims, dtype="int")
        min_pad = np.minimum(cropping, min_pad)

    # perform segmentation
    if len(path_images) <= 10:
        loop_info = utils.LoopInfo(len(path_images), 1, "predicting", True)
    else:
        loop_info = utils.LoopInfo(len(path_images), 10, "predicting", True)
    for i in range(len(path_images)):
        if verbose:
            loop_info.update(i)

        # compute segmentation only if needed
        if compute[i]:

            # preprocessing
            image, mask, aff, h, im_res, shape, pad_idx, crop_idx = preprocess(
                path_image=path_images[i],
                path_mask=path_masks[i],
                n_levels=n_levels,
                crop=cropping,
                min_pad=min_pad,
            )

            # prediction
            post_patch = net.predict([image, mask])

            # postprocessing
            seg, posteriors, volumes = postprocess(
                post_patch=post_patch,
                mask=mask,
                shape=shape,
                pad_idx=pad_idx,
                crop_idx=crop_idx,
                n_dims=n_dims,
                labels_segmentation=labels_segmentation,
                strict_masking=strict_masking,
                keep_biggest_component=keep_biggest_component,
                aff=aff,
                im_res=im_res,
            )

            # write results to disk
            utils.save_volume(seg, aff, h, path_segmentations[i], dtype="int32")
            if path_posteriors[i] is not None:
                if n_channels > 1:
                    posteriors = utils.add_axis(posteriors, axis=[0, -1])
                utils.save_volume(
                    posteriors, aff, h, path_posteriors[i], dtype="float32"
                )

            # compute volumes
            if path_volumes[i] is not None:
                row = [os.path.basename(path_images[i]).replace(".nii.gz", "")] + [
                    str(vol) for vol in volumes
                ]
                write_csv(
                    path_volumes[i],
                    row,
                    unique_vol_file,
                    labels_segmentation,
                    names_segmentation,
                )

    # evaluate
    if gt_folder is not None:

        # find path where segmentations are saved evaluation folder, and get labels on which to evaluate
        eval_folder = os.path.dirname(path_segmentations[0])
        if evaluation_labels is None:
            evaluation_labels = labels_segmentation

        # set path of result arrays for surface distance if necessary
        if compute_distances:
            path_hausdorff = os.path.join(eval_folder, "hausdorff.npy")
            path_hausdorff_99 = os.path.join(eval_folder, "hausdorff_99.npy")
            path_hausdorff_95 = os.path.join(eval_folder, "hausdorff_95.npy")
            path_mean_distance = os.path.join(eval_folder, "mean_distance.npy")
        else:
            path_hausdorff = path_hausdorff_99 = path_hausdorff_95 = (
                path_mean_distance
            ) = None

        # compute evaluation metrics
        evaluate.evaluation(
            gt_folder,
            eval_folder,
            evaluation_labels,
            path_dice=os.path.join(eval_folder, "dice.npy"),
            path_hausdorff=path_hausdorff,
            path_hausdorff_99=path_hausdorff_99,
            path_hausdorff_95=path_hausdorff_95,
            path_mean_distance=path_mean_distance,
            list_incorrect_labels=list_incorrect_labels,
            list_correct_labels=list_correct_labels,
            recompute=recompute,
            verbose=verbose,
        )


def prepare_output_files(
    path_images, path_masks, out_seg, out_posteriors, out_volumes, recompute
):

    # check inputs
    assert path_images is not None, "please specify an input file/folder (--i)"
    assert path_masks is not None, "please specify an input file/folder (--i)"
    assert out_seg is not None, "please specify an output file/folder (--o)"

    # convert path to absolute paths
    path_images = os.path.abspath(path_images)
    path_masks = os.path.abspath(path_masks)
    basename = os.path.basename(path_images)
    out_seg = os.path.abspath(out_seg)
    out_posteriors = (
        os.path.abspath(out_posteriors)
        if (out_posteriors is not None)
        else out_posteriors
    )
    out_volumes = (
        os.path.abspath(out_volumes) if (out_volumes is not None) else out_volumes
    )

    # path_images is a text file
    if basename[-4:] == ".txt":

        # input images
        if not os.path.isfile(path_images):
            raise Exception(
                "provided text file containing paths of input images does not exist"
                % path_images
            )
        with open(path_images, "r") as f:
            path_images = [
                line.replace("\n", "") for line in f.readlines() if line != "\n"
            ]

        # masks
        if not os.path.isfile(path_masks):
            raise Exception(
                "provided text file containing paths of input images does not exist"
                % path_masks
            )
        with open(path_masks, "r") as f:
            path_masks = [
                line.replace("\n", "") for line in f.readlines() if line != "\n"
            ]

        # define helper to deal with outputs
        def text_helper(path, name):
            if path is not None:
                assert path[-4:] == ".txt", (
                    "if path_images given as text file, so must be %s" % name
                )
                with open(path, "r") as ff:
                    path = [
                        line.replace("\n", "")
                        for line in ff.readlines()
                        if line != "\n"
                    ]
                recompute_files = [not os.path.isfile(p) for p in path]
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            unique_file = False
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = text_helper(out_seg, "path_segmentations")
        out_posteriors, recompute_post, _ = text_helper(
            out_posteriors, "path_posteriors"
        )
        out_volumes, recompute_volume, unique_volume_file = text_helper(
            out_volumes, "path_volume"
        )

    # path_images is a folder
    elif (
        (".nii.gz" not in basename)
        & (".nii" not in basename)
        & (".mgz" not in basename)
        & (".npz" not in basename)
    ):

        # input images
        if os.path.isfile(path_images):
            raise Exception(
                "Extension not supported for %s, only use: nii.gz, .nii, .mgz, or .npz"
                % path_images
            )
        path_images = utils.list_images_in_folder(path_images)

        # masks
        if os.path.isfile(path_masks):
            raise Exception(
                "Extension not supported for %s, only use: nii.gz, .nii, .mgz, or .npz"
                % path_masks
            )
        path_masks = utils.list_images_in_folder(path_masks)

        # define helper to deal with outputs
        def helper_dir(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    path = [path] * len(path_images)
                    recompute_files = [True] * len(path_images)
                    unique_file = True
                else:
                    if (
                        (path[-7:] == ".nii.gz")
                        | (path[-4:] == ".nii")
                        | (path[-4:] == ".mgz")
                        | (path[-4:] == ".npz")
                    ):
                        raise Exception("Output FOLDER had a FILE extension" % path)
                    path = [
                        os.path.join(path, os.path.basename(p)) for p in path_images
                    ]
                    path = [p.replace(".nii", "_%s.nii" % suffix) for p in path]
                    path = [p.replace(".mgz", "_%s.mgz" % suffix) for p in path]
                    path = [p.replace(".npz", "_%s.npz" % suffix) for p in path]
                    recompute_files = [not os.path.isfile(p) for p in path]
                utils.mkdir(os.path.dirname(path[0]))
            else:
                path = [None] * len(path_images)
                recompute_files = [False] * len(path_images)
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_dir(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_dir(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_dir(
            out_volumes, "path_volumes", "csv", ""
        )

    # path_images is an image
    else:

        # input images
        assert os.path.isfile(path_images), (
            "file does not exist: %s \n"
            "please make sure the path and the extension are correct" % path_images
        )
        path_images = [path_images]

        # masks
        assert os.path.isfile(path_masks), (
            "file does not exist: %s \n"
            "please make sure the path and the extension are correct" % path_masks
        )
        path_masks = [path_masks]

        # define helper to deal with outputs
        def helper_im(path, name, file_type, suffix):
            unique_file = False
            if path is not None:
                assert path[-4:] != ".txt", (
                    "%s can only be given as text file when path_images is." % name
                )
                if file_type == "csv":
                    if path[-4:] != ".csv":
                        print(
                            "%s provided without csv extension. Adding csv extension."
                            % name
                        )
                        path += ".csv"
                    recompute_files = [True]
                    unique_file = True
                else:
                    if (
                        (".nii.gz" not in path)
                        & (".nii" not in path)
                        & (".mgz" not in path)
                        & (".npz" not in path)
                    ):
                        file_name = os.path.basename(path_images[0]).replace(
                            ".nii", "_%s.nii" % suffix
                        )
                        file_name = file_name.replace(".mgz", "_%s.mgz" % suffix)
                        file_name = file_name.replace(".npz", "_%s.npz" % suffix)
                        path = os.path.join(path, file_name)
                    recompute_files = [not os.path.isfile(path)]
                utils.mkdir(os.path.dirname(path))
            else:
                recompute_files = [False]
            path = [path]
            return path, recompute_files, unique_file

        # use helper on all outputs
        out_seg, recompute_seg, _ = helper_im(
            out_seg, "path_segmentations", "", "synthseg"
        )
        out_posteriors, recompute_post, _ = helper_im(
            out_posteriors, "path_posteriors", "", "posteriors"
        )
        out_volumes, recompute_volume, unique_volume_file = helper_im(
            out_volumes, "path_volumes", "csv", ""
        )

    recompute_list = [
        recompute | re_seg | re_post | re_vol
        for (re_seg, re_post, re_vol) in zip(
            recompute_seg, recompute_post, recompute_volume
        )
    ]

    return (
        path_images,
        path_masks,
        out_seg,
        out_posteriors,
        out_volumes,
        recompute_list,
        unique_volume_file,
    )


def preprocess(path_image, path_mask, n_levels, crop=None, min_pad=None):

    # read image and corresponding info
    im, _, aff, n_dims, n_channels, h, im_res = utils.get_volume_info(path_image, True)
    mask = utils.load_volume(path_mask, True)

    # align image
    im = edit_volumes.align_volume_to_ref(
        im, aff, aff_ref=np.eye(4), n_dims=n_dims, return_copy=False
    )
    mask = edit_volumes.align_volume_to_ref(
        mask, aff, aff_ref=np.eye(4), n_dims=n_dims, return_copy=False
    )
    shape = list(im.shape[:n_dims])

    # crop image if necessary
    if crop is not None:
        crop = utils.reformat_to_list(crop, length=n_dims, dtype="int")
        crop_shape = [
            utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
            for s in crop
        ]
        im, crop_idx = edit_volumes.crop_volume(
            im, cropping_shape=crop_shape, return_crop_idx=True
        )
        mask = edit_volumes.crop_volume_with_idx(mask, crop_idx, n_dims=n_dims)
    else:
        crop_idx = None

    # normalise image
    if n_channels == 1:
        im = edit_volumes.rescale_volume(
            im, new_min=0.0, new_max=1.0, min_percentile=0.5, max_percentile=99.5
        )
    else:
        for i in range(im.shape[-1]):
            im[..., i] = edit_volumes.rescale_volume(
                im[..., i],
                new_min=0.0,
                new_max=1.0,
                min_percentile=0.5,
                max_percentile=99.5,
            )

    # pad image
    input_shape = im.shape[:n_dims]
    pad_shape = [
        utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
        for s in input_shape
    ]
    if (
        min_pad is not None
    ):  # in SynthSeg predict use crop flag and then if used do min_pad=crop else min_pad = 192
        min_pad = utils.reformat_to_list(min_pad, length=n_dims, dtype="int")
        min_pad = [
            utils.find_closest_number_divisible_by_m(s, 2**n_levels, "higher")
            for s in min_pad
        ]
        pad_shape = np.maximum(pad_shape, min_pad)
    im, pad_idx = edit_volumes.pad_volume(
        im, padding_shape=pad_shape, return_pad_idx=True
    )
    mask = edit_volumes.pad_volume(mask, padding_shape=pad_shape)

    # add batch and channel axes
    im = utils.add_axis(im) if n_channels > 1 else utils.add_axis(im, axis=[0, -1])
    mask = utils.add_axis(
        mask, axis=0
    )  # channel axis will be added later when computing one-hot

    return im, mask, aff, h, im_res, shape, pad_idx, crop_idx


def build_model(
    path_model,
    input_shape,
    labels_segmentation,
    n_levels,
    nb_conv_per_level,
    conv_size,
    unet_feat_count,
    feat_multiplier,
    activation,
    sigma_smoothing,
    mask_labels,
):

    assert os.path.isfile(path_model), "The provided model path does not exist."

    # get labels
    n_labels_seg = len(labels_segmentation)

    # one-hot encoding of the input prediction as the network expects soft probabilities
    input_image = KL.Input(input_shape)
    input_labels = KL.Input(input_shape[:-1], dtype="int32")
    labels = KL.Lambda(
        lambda x: tf.one_hot(tf.cast(x, "int32"), depth=len(mask_labels), axis=-1)
    )(input_labels)
    image = KL.Lambda(lambda x: tf.cast(tf.concat(x, axis=-1), "float32"))(
        [input_image, labels]
    )
    net = Model(inputs=[input_image, input_labels], outputs=image)

    # build UNet
    net = nrn_models.unet(
        input_model=net,
        input_shape=input_shape,
        nb_labels=n_labels_seg,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
    )
    net.load_weights(path_model, by_name=True)

    # smooth posteriors if specified
    if sigma_smoothing > 0:
        last_tensor = net.output
        last_tensor._keras_shape = tuple(last_tensor.get_shape().as_list())
        last_tensor = layers.GaussianBlur(sigma=sigma_smoothing)(last_tensor)
        net = Model(inputs=net.inputs, outputs=last_tensor)

    return net


def postprocess(
    post_patch,
    mask,
    shape,
    pad_idx,
    crop_idx,
    n_dims,
    labels_segmentation,
    strict_masking,
    keep_biggest_component,
    aff,
    im_res,
):

    # get posteriors
    post_patch = np.squeeze(post_patch)
    mask = np.squeeze(mask)

    # reset posteriors of background to 1 outside mask and to 0 inside
    if strict_masking:
        post_patch[..., 0] = np.ones_like(post_patch[..., 0])
        post_patch[..., 0] = edit_volumes.mask_volume(
            post_patch[..., 0], mask=mask < 0.1, return_copy=False
        )
    # keep biggest connected component (use it with smoothing!)
    elif keep_biggest_component:
        tmp_post_patch = post_patch[..., 1:]
        post_patch_mask = np.sum(tmp_post_patch, axis=-1) > 0.25
        post_patch_mask = edit_volumes.get_largest_connected_component(post_patch_mask)
        post_patch_mask = np.stack(
            [post_patch_mask] * tmp_post_patch.shape[-1], axis=-1
        )
        tmp_post_patch = edit_volumes.mask_volume(
            tmp_post_patch, mask=post_patch_mask, return_copy=False
        )
        post_patch[..., 1:] = tmp_post_patch

    # normalise posteriors and get hard segmentation
    if strict_masking | keep_biggest_component:
        post_patch /= np.sum(post_patch, axis=-1)[..., np.newaxis]
    seg_patch = labels_segmentation[post_patch.argmax(-1).astype("int32")].astype(
        "int32"
    )

    # paste patches back to matrix of original image size
    seg_patch = edit_volumes.crop_volume_with_idx(
        seg_patch, pad_idx, n_dims=n_dims, return_copy=False
    )
    post_patch = edit_volumes.crop_volume_with_idx(
        post_patch, pad_idx, n_dims=n_dims, return_copy=False
    )
    if crop_idx is not None:
        # we need to go through this because of the posteriors of the background, otherwise pad_volume would work
        seg = np.zeros(shape=shape, dtype="int32")
        posteriors = np.zeros(shape=[*shape, labels_segmentation.shape[0]])
        posteriors[..., 0] = np.ones(shape)  # place background around patch
        if n_dims == 2:
            seg[crop_idx[0] : crop_idx[2], crop_idx[1] : crop_idx[3]] = seg_patch
            posteriors[crop_idx[0] : crop_idx[2], crop_idx[1] : crop_idx[3], :] = (
                post_patch
            )
        elif n_dims == 3:
            seg[
                crop_idx[0] : crop_idx[3],
                crop_idx[1] : crop_idx[4],
                crop_idx[2] : crop_idx[5],
            ] = seg_patch
            posteriors[
                crop_idx[0] : crop_idx[3],
                crop_idx[1] : crop_idx[4],
                crop_idx[2] : crop_idx[5],
                :,
            ] = post_patch
    else:
        seg = seg_patch
        posteriors = post_patch

    # align prediction back to first orientation
    seg = edit_volumes.align_volume_to_ref(
        seg, aff=np.eye(4), aff_ref=aff, n_dims=n_dims, return_copy=False
    )
    posteriors = edit_volumes.align_volume_to_ref(
        posteriors, np.eye(4), aff_ref=aff, n_dims=n_dims, return_copy=False
    )

    # compute volumes
    volumes = np.sum(
        posteriors[..., 1:], axis=tuple(range(0, len(posteriors.shape) - 1))
    )
    volumes = np.around(volumes * np.prod(im_res), 3)

    return seg, posteriors, volumes



"""
apply_warp - Image registration transformation application

Part of the micaflow processing pipeline for neuroimaging data.

This module applies spatial transformations to register images from one space to another
using both affine and non-linear (warp field) transformations. It's commonly used to:
- Transform subject images to a standard space (e.g., MNI152)
- Register images across modalities (e.g., T1w to FLAIR)
- Apply previously calculated transformations to derived images (e.g., segmentations)

The module leverages ANTsPy to apply the transformations in the correct order (warp
field first, then affine) to achieve accurate spatial registration.

API Usage:
---------
micaflow apply_warp
    --moving <path/to/source_image.nii.gz>
    --reference <path/to/target_space.nii.gz>
    --affine <path/to/transform.mat>
    --warp <path/to/warpfield.nii.gz>
    [--output <path/to/registered_image.nii.gz>]

Python Usage:
-----------
>>> import ants
>>> from micaflow.scripts.apply_warp import apply_warp
>>> moving_img = ants.image_read("subject_t1w.nii.gz")
>>> reference_img = ants.image_read("mni152.nii.gz")
>>> apply_warp(
...     moving_img=moving_img,
...     reference_img=reference_img,
...     affine_file="transform.mat",
...     warp_file="warpfield.nii.gz",
...     out_file="registered_t1w.nii.gz"
... )

References:
----------
1. Avants BB, Tustison NJ, Song G, et al. A reproducible evaluation of ANTs
   similarity metric performance in brain image registration. NeuroImage.
   2011;54(3):2033-2044. doi:10.1016/j.neuroimage.2010.09.025
"""

import ants
import argparse
import sys
from colorama import init, Fore, Style

init()


def print_help():
    """Print a help message with examples."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                        APPLY WARP                              ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}
    
    This script applies both an affine transformation and a warp field to
    register a moving image to a reference space.
    
    {CYAN}{BOLD}────────────────────────── REQUIRED ARGUMENTS ──────────────────────────{RESET}
      {YELLOW}--moving{RESET}     : Path to the input image to be warped (.nii.gz)
      {YELLOW}--fixed{RESET}      : Path to the target/reference image (.nii.gz)
      {YELLOW}--affine{RESET}     : Path to the affine transformation file (.mat)
      {YELLOW}--warp{RESET}       : Path to the warp field (.nii.gz)
    
    {CYAN}{BOLD}────────────────────────── OPTIONAL ARGUMENTS ──────────────────────────{RESET}
      {YELLOW}--output{RESET}     : Output path for the warped image (default: warped_image.nii.gz)
      {YELLOW}--inverse{RESET}    : Reverse the transform order (apply warpfield first, then affine)
    
    {CYAN}{BOLD}────────────────────────── EXAMPLE USAGE ──────────────────────────{RESET}
    
    {BLUE}# Apply warp transformation{RESET}
    micaflow {GREEN}apply_warp{RESET} {YELLOW}--moving{RESET} subject_t1w.nii.gz {YELLOW}--fixed{RESET} mni152.nii.gz \\
      {YELLOW}--affine{RESET} transform.mat {YELLOW}--warp{RESET} warpfield.nii.gz {YELLOW}--output{RESET} registered_t1w.nii.gz
    """

    print(help_text)


def apply_warp(
    moving_img, reference_img, affine_file, warp_file, out_file, interpolation="linear", inverse=False
):
    """Apply an affine transform and a warp field to a moving image.

    This function takes a moving image and applies both an affine transformation
    and a nonlinear warp field to register it to a reference image space. The
    transformation is applied using ANTsPy's apply_transforms function with the
    appropriate transform order.

    Parameters
    ----------
    moving_file : str
        Path to the moving image that will be transformed (.nii.gz).
    reference_file : str
        Path to the reference/fixed image that defines the target space (.nii.gz).
    affine_file : str
        Path to the affine transformation file (.mat).
    warp_file : str
        Path to the nonlinear warp field (.nii.gz).
    out_file : str
        Path where the transformed image will be saved.
    interpolation : str, optional
        Interpolation method to use for the transformation. Default is 'linear'.
        Other options include 'nearestNeighbor', 'multiLabel', etc.

    Returns
    -------
    None
        The function saves the transformed image to the specified output path
        but does not return any values.

    Notes
    -----
    The order of transforms matters: the warp field is applied first, followed
    by the affine transformation. This is the standard order in ANTs for
    composite transformations.
    """

    # The order of transforms in transformlist matters (last Transform will be applied first).
    # Usually you put the nonlinear warp first, then the affine:
    if inverse:
        transformed = ants.apply_transforms(
            fixed=reference_img,
            moving=moving_img,
            transformlist=[affine_file, warp_file],
            interpolator=interpolation,
        )
    else:
        transformed = ants.apply_transforms(
            fixed=reference_img,
            moving=moving_img,
            transformlist=[warp_file, affine_file],
            interpolator=interpolation,
        )

    # Save the transformed image
    ants.image_write(transformed, out_file)
    print(f"Saved warped image as {out_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Apply an affine (.mat) and a warp field (.nii.gz) to an image using ANTsPy."
    )
    parser.add_argument(
        "--moving", required=True, help="Path to the moving image (.nii.gz)."
    )
    parser.add_argument(
        "--fixed", required=True, help="Path to the fixed/reference image (.nii.gz)."
    )
    parser.add_argument(
        "--output", default="warped_image.nii.gz", help="Output warped image filename."
    )
    parser.add_argument(
        "--affine", required=True, help="Path to the affine transform (.mat)."
    )
    parser.add_argument(
        "--warp", required=True, help="Path to the warp field (.nii.gz)."
    )
    parser.add_argument(
        "--interpolation",
        default="linear",
        help="Interpolation method (default: linear).",
    )
    parser.add_argument(
        "--inverse",
        action="store_true",
        help="Reverse the transform order (apply warpfield first, then affine)"
    )
    args = parser.parse_args()

    moving_img = ants.image_read(args.moving)
    reference_img = ants.image_read(args.fixed)

    apply_warp(
        moving_img,
        reference_img,
        args.affine,
        args.warp,
        args.output,
        args.interpolation,
        inverse=args.inverse
    )


if __name__ == "__main__":
    main()



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np

try:
    from scipy.stats import median_absolute_deviation
except ImportError:
    from scipy.stats import median_abs_deviation as median_absolute_deviation


# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import edit_volumes


def sample_intensity_stats_from_image(
    image, segmentation, labels_list, classes_list=None, keep_strictly_positive=True
):
    """This function takes an image and corresponding segmentation as inputs. It estimates the mean and std intensity
    for all specified label values. Labels can share the same statistics by being regrouped into K classes.
    :param image: image from which to evaluate mean intensity and std deviation.
    :param segmentation: segmentation of the input image. Must have the same size as image.
    :param labels_list: list of labels for which to evaluate mean and std intensity.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    :param classes_list: (optional) enables to regroup structures into classes of similar intensity statistics.
    Intensities associated to regrouped labels will thus contribute to the same Gaussian during statistics estimation.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    It should have the same length as labels_list, and contain values between 0 and K-1, where K is the total number of
    classes. Default is all labels have different classes (K=len(labels_list)).
    :param keep_strictly_positive: (optional) whether to only keep strictly positive intensity values when
    computing stats. This doesn't apply to the first label in label_list (or class if class_list is provided), for
    which we keep positive and zero values, as we consider it to be the background label.
    :return: a numpy array of size (2, K), the first row being the mean intensity for each structure,
    and the second being the median absolute deviation (robust estimation of std).
    """

    # reformat labels and classes
    labels_list = np.array(
        utils.reformat_to_list(labels_list, load_as_numpy=True, dtype="int")
    )
    if classes_list is not None:
        classes_list = np.array(
            utils.reformat_to_list(classes_list, load_as_numpy=True, dtype="int")
        )
    else:
        classes_list = np.arange(labels_list.shape[0])
    assert len(classes_list) == len(
        labels_list
    ), "labels and classes lists should have the same length"

    # get unique classes
    unique_classes, unique_indices = np.unique(classes_list, return_index=True)
    n_classes = len(unique_classes)
    if not np.array_equal(unique_classes, np.arange(n_classes)):
        raise ValueError(
            "classes_list should only contain values between 0 and K-1, "
            "where K is the total number of classes. Here K = %d" % n_classes
        )

    # compute mean/std of specified classes
    means = np.zeros(n_classes)
    stds = np.zeros(n_classes)
    for idx, tmp_class in enumerate(unique_classes):

        # get list of all intensity values for the current class
        class_labels = labels_list[classes_list == tmp_class]
        intensities = np.empty(0)
        for label in class_labels:
            tmp_intensities = image[segmentation == label]
            intensities = np.concatenate([intensities, tmp_intensities])
        if tmp_class:  # i.e. if not background
            if keep_strictly_positive:
                intensities = intensities[intensities > 0]

        # compute stats for class and put them to the location of corresponding label values
        if len(intensities) != 0:
            means[idx] = np.nanmedian(intensities)
            stds[idx] = median_absolute_deviation(intensities, nan_policy="omit")

    return np.stack([means, stds])


def sample_intensity_stats_from_single_dataset(
    image_dir, labels_dir, labels_list, classes_list=None, max_channel=3, rescale=True
):
    """This function aims at estimating the intensity distributions of K different structure types from a set of images.
    The distribution of each structure type is modelled as a Gaussian, parametrised by a mean and a standard deviation.
    Because the intensity distribution of structures can vary across images, we additionally use Gaussian priors for the
    parameters of each Gaussian distribution. Therefore, the intensity distribution of each structure type is described
    by 4 parameters: a mean/std for the mean intensity, and a mean/std for the std deviation.
    This function uses a set of images along with corresponding segmentations to estimate the 4*K parameters.
    Structures can share the same statistics by being regrouped into classes of similar structure types.
    Images can be multi-modal (n_channels), in which case different statistics are estimated for each modality.
    :param image_dir: path of directory with images to estimate the intensity distribution
    :param labels_dir: path of directory with segmentation of input images.
    They are matched with images by sorting order.
    :param labels_list: list of labels for which to evaluate mean and std intensity.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    :param classes_list: (optional) enables to regroup structures into classes of similar intensity statistics.
    Intensities associated to regrouped labels will thus contribute to the same Gaussian during statistics estimation.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    It should have the same length as labels_list, and contain values between 0 and K-1, where K is the total number of
    classes. Default is all labels have different classes (K=len(labels_list)).
    :param max_channel: (optional) maximum number of channels to consider if the data is multi-spectral. Default is 3.
    :param rescale: (optional) whether to rescale images between 0 and 255 before intensity estimation
    :return: 2 numpy arrays of size (2*n_channels, K), one with the evaluated means/std for the mean
    intensity, and one for the mean/std for the standard deviation.
    Each block of two rows correspond to a different modality (channel). For each block of two rows, the first row
    represents the mean, and the second represents the std.
    """

    # list files
    path_images = utils.list_images_in_folder(image_dir)
    path_labels = utils.list_images_in_folder(labels_dir)
    assert len(path_images) == len(
        path_labels
    ), "image and labels folders do not have the same number of files"

    # reformat list labels and classes
    labels_list = np.array(
        utils.reformat_to_list(labels_list, load_as_numpy=True, dtype="int")
    )
    if classes_list is not None:
        classes_list = np.array(
            utils.reformat_to_list(classes_list, load_as_numpy=True, dtype="int")
        )
    else:
        classes_list = np.arange(labels_list.shape[0])
    assert len(classes_list) == len(
        labels_list
    ), "labels and classes lists should have the same length"

    # get unique classes
    unique_classes, unique_indices = np.unique(classes_list, return_index=True)
    n_classes = len(unique_classes)
    if not np.array_equal(unique_classes, np.arange(n_classes)):
        raise ValueError(
            "classes_list should only contain values between 0 and K-1, "
            "where K is the total number of classes. Here K = %d" % n_classes
        )

    # initialise result arrays
    n_dims, n_channels = utils.get_dims(
        utils.load_volume(path_images[0]).shape, max_channels=max_channel
    )
    means = np.zeros((len(path_images), n_classes, n_channels))
    stds = np.zeros((len(path_images), n_classes, n_channels))

    # loop over images
    loop_info = utils.LoopInfo(len(path_images), 10, "estimating", print_time=True)
    for idx, (path_im, path_la) in enumerate(zip(path_images, path_labels)):
        loop_info.update(idx)

        # load image and label map
        image = utils.load_volume(path_im)
        la = utils.load_volume(path_la)
        if n_channels == 1:
            image = utils.add_axis(image, -1)

        # loop over channels
        for channel in range(n_channels):
            im = image[..., channel]
            if rescale:
                im = edit_volumes.rescale_volume(im)
            stats = sample_intensity_stats_from_image(
                im, la, labels_list, classes_list=classes_list
            )
            means[idx, :, channel] = stats[0, :]
            stds[idx, :, channel] = stats[1, :]

    # compute prior parameters for mean/std
    mean_means = np.mean(means, axis=0)
    std_means = np.std(means, axis=0)
    mean_stds = np.mean(stds, axis=0)
    std_stds = np.std(stds, axis=0)

    # regroup prior parameters in two different arrays: one for the mean and one for the std
    prior_means = np.zeros((2 * n_channels, n_classes))
    prior_stds = np.zeros((2 * n_channels, n_classes))
    for channel in range(n_channels):
        prior_means[2 * channel, :] = mean_means[:, channel]
        prior_means[2 * channel + 1, :] = std_means[:, channel]
        prior_stds[2 * channel, :] = mean_stds[:, channel]
        prior_stds[2 * channel + 1, :] = std_stds[:, channel]

    return prior_means, prior_stds


def build_intensity_stats(
    list_image_dir,
    list_labels_dir,
    result_dir,
    estimation_labels,
    estimation_classes=None,
    max_channel=3,
    rescale=True,
):
    """This function aims at estimating the intensity distributions of K different structure types from a set of images.
    The distribution of each structure type is modelled as a Gaussian, parametrised by a mean and a standard deviation.
    Because the intensity distribution of structures can vary across images, we additionally use Gaussian priors for the
    parameters of each Gaussian distribution. Therefore, the intensity distribution of each structure type is described
    by 4 parameters: a mean/std for the mean intensity, and a mean/std for the std deviation.
    This function uses a set of images along with corresponding segmentations to estimate the 4*K parameters.
    Additionally, it can estimate the 4*K parameters for several image datasets, that we call here n_datasets.
    This function writes 2 numpy arrays of size (2*n_datasets, K), one with the evaluated means/std for the mean
    intensities, and one for the mean/std for the standard deviations.
    In these arrays, each block of two rows refer to a different dataset.
    Within each block of two rows, the first row represents the mean, and the second represents the std.
    :param list_image_dir: path of folders with images for intensity distribution estimation.
    Can be the path of single directory (n_datasets=1), or a list of folders, each being a separate dataset.
    Images can be multimodal, in which case each modality is treated as a different dataset, i.e. each modality will
    have a separate block (of size (2, K)) in the result arrays.
    :param list_labels_dir: path of folders with label maps corresponding to input images.
    If list_image_dir is a list of several folders, list_labels_dir can either be a list of folders (one for each image
    folder), or the path to a single folder, which will be used for all datasets.
    If a dataset has multi-modal images, the same label map is applied to all modalities.
    :param result_dir: path of directory where estimated priors will be writen.
    :param estimation_labels: labels to estimate intensity statistics from.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    :param estimation_classes: (optional) enables to regroup structures into classes of similar intensity statistics.
    Intensities associated to regrouped labels will thus contribute to the same Gaussian during statistics estimation.
    Can be a sequence, a 1d numpy array, or the path to a 1d numpy array.
    It should have the same length as labels_list, and contain values between 0 and K-1, where K is the total number of
    classes. Default is all labels have different classes (K=len(estimation_labels)).
    :param max_channel: (optional) maximum number of channels to consider if the data is multi-spectral. Default is 3.
    :param rescale: (optional) whether to rescale images between 0 and 255 before intensity estimation
    """

    # handle results directories
    utils.mkdir(result_dir)

    # reformat image/labels dir into lists
    list_image_dir = utils.reformat_to_list(list_image_dir)
    list_labels_dir = utils.reformat_to_list(
        list_labels_dir, length=len(list_image_dir)
    )

    # reformat list estimation labels and classes
    estimation_labels = np.array(
        utils.reformat_to_list(estimation_labels, load_as_numpy=True, dtype="int")
    )
    if estimation_classes is not None:
        estimation_classes = np.array(
            utils.reformat_to_list(estimation_classes, load_as_numpy=True, dtype="int")
        )
    else:
        estimation_classes = np.arange(estimation_labels.shape[0])
    assert len(estimation_classes) == len(
        estimation_labels
    ), "estimation labels and classes should be of same length"

    # get unique classes
    unique_estimation_classes, unique_indices = np.unique(
        estimation_classes, return_index=True
    )
    n_classes = len(unique_estimation_classes)
    if not np.array_equal(unique_estimation_classes, np.arange(n_classes)):
        raise ValueError(
            "estimation_classes should only contain values between 0 and N-1, "
            "where K is the total number of classes. Here N = %d" % n_classes
        )

    # loop over dataset
    list_datasets_prior_means = list()
    list_datasets_prior_stds = list()
    for image_dir, labels_dir in zip(list_image_dir, list_labels_dir):

        # get prior stats for dataset
        tmp_prior_means, tmp_prior_stds = sample_intensity_stats_from_single_dataset(
            image_dir,
            labels_dir,
            estimation_labels,
            estimation_classes,
            max_channel=max_channel,
            rescale=rescale,
        )

        # add stats arrays to list of datasets-wise statistics
        list_datasets_prior_means.append(tmp_prior_means)
        list_datasets_prior_stds.append(tmp_prior_stds)

    # stack all modalities together
    prior_means = np.concatenate(list_datasets_prior_means, axis=0)
    prior_stds = np.concatenate(list_datasets_prior_stds, axis=0)

    # save files
    np.save(os.path.join(result_dir, "prior_means.npy"), prior_means)
    np.save(os.path.join(result_dir, "prior_stds.npy"), prior_stds)

    return prior_means, prior_stds



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os

# project imports
from lamareg.SynthSeg.predict_denoiser import predict

# third-party imports
from lamareg.ext.lab2im import utils


def validate_training(
    prediction_dir,
    gt_dir,
    models_dir,
    validation_main_dir,
    target_segmentation_labels,
    input_segmentation_labels=None,
    evaluation_labels=None,
    step_eval=1,
    min_pad=None,
    cropping=None,
    topology_classes=None,
    sigma_smoothing=0,
    keep_biggest_component=False,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    skip_n_concatenations=0,
    recompute=True,
):

    # create result folder
    utils.mkdir(validation_main_dir)

    # loop over models
    list_models = utils.list_files(models_dir, expr=["dice", ".h5"], cond_type="and")[
        ::step_eval
    ]
    # list_models = [p for p in list_models if int(os.path.basename(p)[-6:-3]) % 2 == 0]
    loop_info = utils.LoopInfo(len(list_models), 1, "validating", True)
    for model_idx, path_model in enumerate(list_models):

        # build names and create folders
        model_val_dir = os.path.join(
            validation_main_dir, os.path.basename(path_model).replace(".h5", "")
        )
        dice_path = os.path.join(model_val_dir, "dice.npy")
        utils.mkdir(model_val_dir)

        if (not os.path.isfile(dice_path)) | recompute:
            loop_info.update(model_idx)
            predict(
                path_predictions=prediction_dir,
                path_corrections=model_val_dir,
                path_model=path_model,
                target_segmentation_labels=target_segmentation_labels,
                input_segmentation_labels=input_segmentation_labels,
                min_pad=min_pad,
                cropping=cropping,
                topology_classes=topology_classes,
                sigma_smoothing=sigma_smoothing,
                keep_biggest_component=keep_biggest_component,
                n_levels=n_levels,
                nb_conv_per_level=nb_conv_per_level,
                conv_size=conv_size,
                unet_feat_count=unet_feat_count,
                feat_multiplier=feat_multiplier,
                activation=activation,
                skip_n_concatenations=skip_n_concatenations,
                gt_folder=gt_dir,
                evaluation_labels=evaluation_labels,
                recompute=recompute,
                verbose=False,
            )



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np
import tensorflow as tf
import keras.layers as KL
from keras.models import Model

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.lab2im import edit_tensors as l2i_et
from lamareg.ext.lab2im.edit_volumes import get_ras_axes


def labels_to_image_model(
    labels_shape,
    n_channels,
    generation_labels,
    output_labels,
    n_neutral_labels,
    atlas_res,
    target_res,
    output_shape=None,
    output_div_by_n=None,
    flipping=True,
    aff=None,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_scale=0.0625,
    randomise_res=False,
    max_res_iso=4.0,
    max_res_aniso=8.0,
    data_res=None,
    thickness=None,
    bias_field_std=0.5,
    bias_scale=0.025,
    return_gradients=False,
):
    """
    This function builds a keras/tensorflow model to generate images from provided label maps.
    The images are generated by sampling a Gaussian Mixture Model (of given parameters), conditioned on the label map.
    The model will take as inputs:
        -a label map
        -a vector containing the means of the Gaussian Mixture Model for each label,
        -a vector containing the standard deviations of the Gaussian Mixture Model for each label,
    The model returns:
        -the generated image normalised between 0 and 1.
        -the corresponding label map, with only the labels present in output_labels (the other are reset to zero).
    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.
    :param labels_shape: shape of the input label maps. Can be a sequence or a 1d numpy array.
    :param n_channels: number of channels to be synthesised.
    :param generation_labels: (optional) list of all possible label values in the input label maps.
    Default is None, where the label values are directly gotten from the provided label maps.
    If not None, can be a sequence or a 1d numpy array. It should be organised as follows: background label first, then
    non-sided labels (e.g. CSF, brainstem, etc.), then all the structures of the same hemisphere (can be left or right),
    and finally all the corresponding contralateral structures (in the same order).
    :param output_labels: (optional) list of the same length as generation_labels to indicate which values to use in the
    label maps returned by this model, i.e. all occurrences of generation_labels[i] in the input label maps will be
    converted to output_labels[i] in the returned label maps. Examples:
    Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
    Set output_labels[i]=generation_labels[i] to keep the value generation_labels[i] in the returned maps.
    Can be a list or a 1d numpy array. By default output_labels is equal to generation_labels.
    :param n_neutral_labels: number of non-sided generation labels.
    :param atlas_res: resolution of the input label maps.
    Can be a number (isotropic resolution), a sequence, or a 1d numpy array.
    :param target_res: target resolution of the generated images and corresponding label maps.
    Can be a number (isotropic resolution), a sequence, or a 1d numpy array.
    :param output_shape: (optional) desired shape of the output image, obtained by randomly cropping the generated image
    Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
    Default is None, where no cropping is performed.
    :param output_div_by_n: (optional) forces the output shape to be divisible by this value. It overwrites output_shape
    if necessary. Can be an integer (same size in all dimensions), a sequence, or a 1d numpy array.
    :param flipping: (optional) whether to introduce right/left random flipping
    :param aff: (optional) example of an (n_dims+1)x(n_dims+1) affine matrix of one of the input label map.
    Used to find brain's right/left axis. Should be given if flipping is True.
    :param scaling_bounds: (optional) range of the random scaling to apply at each mini-batch. The scaling factor for
    each dimension is sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    [1-scaling_bounds, 1+scaling_bounds] for each dimension.
    2) a sequence, in which case the scaling factor is sampled from the uniform distribution of bounds
    (1-scaling_bounds[i], 1+scaling_bounds[i]) for the i-th dimension.
    3) a numpy array of shape (2, n_dims), in which case the scaling factor is sampled from the uniform distribution
     of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    4) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for cases 1
    and 2, the bounds are centred on 0 rather than 1, i.e. [0+rotation_bounds[i], 0-rotation_bounds[i]].
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param translation_bounds: (optional) same as scaling bounds. Default is translation_bounds = False, but we
    encourage using it when cropping is deactivated (i.e. when output_shape=None in BrainGenerator).
    :param nonlin_std: (optional) Maximum value for the standard deviation of the normal distribution from which we
    sample the first tensor for synthesising the deformation field. Set to 0 if you wish to completely turn the elastic
    deformation off.
    :param nonlin_scale: (optional) if nonlin_std is strictly positive, factor between the shapes of the input
    label maps and the shape of the input non-linear tensor.
    :param randomise_res: (optional) whether to mimic images that would have been 1) acquired at low resolution, and
    2) resampled to high resolution. The low resolution is uniformly resampled at each minibatch from [1mm, 9mm].
    In that process, the images generated by sampling the GMM are 1) blurred at the sampled LR, 2) downsampled at LR,
    and 3) resampled at target_resolution.
    :param max_res_iso: (optional) If randomise_res is True, this enables to control the upper bound of the uniform
    distribution from which we sample the random resolution U(min_res, max_res_iso), where min_res is the resolution of
    the input label maps. Must be a number, and default is 4. Set to None to deactivate it, but if randomise_res is
    True, at least one of max_res_iso or max_res_aniso must be given.
    :param max_res_aniso: If randomise_res is True, this enables to downsample the input volumes to a random LR in
    only 1 (random) direction. This is done by randomly selecting a direction i in the range [0, n_dims-1], and sampling
    a value in the corresponding uniform distribution U(min_res[i], max_res_aniso[i]), where min_res is the resolution
    of the input label maps. Can be a number, a sequence, or a 1d numpy array. Set to None to deactivate it, but if
    randomise_res is True, at least one of max_res_iso or max_res_aniso must be given.
    :param data_res: (optional) specific acquisition resolution to mimic, as opposed to random resolution sampled when
    randomise_res is True. This triggers a blurring to mimic the specified acquisition resolution, but the downsampling
    is optional (see param downsample). Default for data_res is None, where images are slightly blurred.
    If the generated images are uni-modal, data_res can be a number (isotropic acquisition resolution), a sequence, a 1d
    numpy array, or the path to a 1d numpy array. In the multi-modal case, it should be given as a numpy array (or a
    path) of size (n_mod, n_dims), where each row is the acquisition resolution of the corresponding channel.
    :param thickness: (optional) if data_res is provided, we can further specify the slice thickness of the low
    resolution images to mimic. Must be provided in the same format as data_res. Default thickness = data_res.
    :param bias_field_std: (optional) If strictly positive, this triggers the corruption of synthesised images with a
    bias field. It is obtained by sampling a first small tensor from a normal distribution, resizing it to full size,
    and rescaling it to positive values by taking the voxel-wise exponential. bias_field_std designates the std dev of
    the normal distribution from which we sample the first tensor. Set to 0 to deactivate bias field corruption.
    :param bias_scale: (optional) If bias_field_std is strictly positive, this designates the ratio between the
    size of the input label maps and the size of the first sampled tensor for synthesising the bias field.
    :param return_gradients: (optional) whether to return the synthetic image or the magnitude of its spatial gradient
    (computed with Sobel kernels).
    """

    # reformat resolutions
    labels_shape = utils.reformat_to_list(labels_shape)
    n_dims, _ = utils.get_dims(labels_shape)
    atlas_res = utils.reformat_to_n_channels_array(atlas_res, n_dims, n_channels)
    data_res = (
        atlas_res
        if data_res is None
        else utils.reformat_to_n_channels_array(data_res, n_dims, n_channels)
    )
    thickness = (
        data_res
        if thickness is None
        else utils.reformat_to_n_channels_array(thickness, n_dims, n_channels)
    )
    atlas_res = atlas_res[0]
    target_res = (
        atlas_res
        if target_res is None
        else utils.reformat_to_n_channels_array(target_res, n_dims)[0]
    )

    # get shapes
    crop_shape, output_shape = get_shapes(
        labels_shape, output_shape, atlas_res, target_res, output_div_by_n
    )

    # define model inputs
    labels_input = KL.Input(
        shape=labels_shape + [1], name="labels_input", dtype="int32"
    )
    means_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="means_input"
    )
    stds_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="std_devs_input"
    )
    list_inputs = [labels_input, means_input, stds_input]

    # deform labels
    labels = layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        inter_method="nearest",
    )(labels_input)

    # cropping
    if crop_shape != labels_shape:
        labels = layers.RandomCrop(crop_shape)(labels)

    # flipping
    if flipping:
        assert aff is not None, "aff should not be None if flipping is True"
        labels = layers.RandomFlip(
            get_ras_axes(aff, n_dims)[0], True, generation_labels, n_neutral_labels
        )(labels)

    # build synthetic image
    image = layers.SampleConditionalGMM(generation_labels)(
        [labels, means_input, stds_input]
    )

    # apply bias field
    if bias_field_std > 0:
        image = layers.BiasFieldCorruption(bias_field_std, bias_scale, False)(image)

    # intensity augmentation
    image = layers.IntensityAugmentation(
        clip=300, normalise=True, gamma_std=0.5, separate_channels=True
    )(image)

    # loop over channels
    channels = list()
    split = (
        KL.Lambda(lambda x: tf.split(x, [1] * n_channels, axis=-1))(image)
        if (n_channels > 1)
        else [image]
    )
    for i, channel in enumerate(split):

        if randomise_res:
            max_res_iso = np.array(
                utils.reformat_to_list(max_res_iso, length=n_dims, dtype="float")
            )
            max_res_aniso = np.array(
                utils.reformat_to_list(max_res_aniso, length=n_dims, dtype="float")
            )
            max_res = np.maximum(max_res_iso, max_res_aniso)
            resolution, blur_res = layers.SampleResolution(
                atlas_res, max_res_iso, max_res_aniso
            )(means_input)
            sigma = l2i_et.blurring_sigma_for_downsampling(
                atlas_res, resolution, thickness=blur_res
            )
            channel = layers.DynamicGaussianBlur(
                0.75 * max_res / np.array(atlas_res), 1.03
            )([channel, sigma])
            channel = layers.MimicAcquisition(
                atlas_res, atlas_res, output_shape, False
            )([channel, resolution])
            channels.append(channel)

        else:
            sigma = l2i_et.blurring_sigma_for_downsampling(
                atlas_res, data_res[i], thickness=thickness[i]
            )
            channel = layers.GaussianBlur(sigma, 1.03)(channel)
            resolution = KL.Lambda(
                lambda x: tf.convert_to_tensor(data_res[i], dtype="float32")
            )([])
            channel = layers.MimicAcquisition(atlas_res, data_res[i], output_shape)(
                [channel, resolution]
            )
            channels.append(channel)

    # concatenate all channels back
    image = (
        KL.Lambda(lambda x: tf.concat(x, -1))(channels)
        if len(channels) > 1
        else channels[0]
    )

    # compute image gradient
    if return_gradients:
        image = layers.ImageGradients("sobel", True, name="image_gradients")(image)
        image = layers.IntensityAugmentation(clip=10, normalise=True)(image)

    # resample labels at target resolution
    if crop_shape != output_shape:
        labels = l2i_et.resample_tensor(labels, output_shape, interp_method="nearest")

    # map generation labels to segmentation values
    labels = layers.ConvertLabels(
        generation_labels, dest_values=output_labels, name="labels_out"
    )(labels)

    # build model (dummy layer enables to keep the labels when plugging this model to other models)
    image = KL.Lambda(lambda x: x[0], name="image_out")([image, labels])
    brain_model = Model(inputs=list_inputs, outputs=[image, labels])

    return brain_model


def get_shapes(labels_shape, output_shape, atlas_res, target_res, output_div_by_n):

    # reformat resolutions to lists
    atlas_res = utils.reformat_to_list(atlas_res)
    n_dims = len(atlas_res)
    target_res = utils.reformat_to_list(target_res)

    # get resampling factor
    if atlas_res != target_res:
        resample_factor = [atlas_res[i] / float(target_res[i]) for i in range(n_dims)]
    else:
        resample_factor = None

    # output shape specified, need to get cropping shape, and resample shape if necessary
    if output_shape is not None:
        output_shape = utils.reformat_to_list(output_shape, length=n_dims, dtype="int")

        # make sure that output shape is smaller or equal to label shape
        if resample_factor is not None:
            output_shape = [
                min(int(labels_shape[i] * resample_factor[i]), output_shape[i])
                for i in range(n_dims)
            ]
        else:
            output_shape = [
                min(labels_shape[i], output_shape[i]) for i in range(n_dims)
            ]

        # make sure output shape is divisible by output_div_by_n
        if output_div_by_n is not None:
            tmp_shape = [
                utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                for s in output_shape
            ]
            if output_shape != tmp_shape:
                print(
                    "output shape {0} not divisible by {1}, changed to {2}".format(
                        output_shape, output_div_by_n, tmp_shape
                    )
                )
                output_shape = tmp_shape

        # get cropping and resample shape
        if resample_factor is not None:
            cropping_shape = [
                int(np.around(output_shape[i] / resample_factor[i], 0))
                for i in range(n_dims)
            ]
        else:
            cropping_shape = output_shape

    # no output shape specified, so no cropping unless label_shape is not divisible by output_div_by_n
    else:

        # make sure output shape is divisible by output_div_by_n
        if output_div_by_n is not None:

            # if resampling, get the potential output_shape and check if it is divisible by n
            if resample_factor is not None:
                output_shape = [
                    int(labels_shape[i] * resample_factor[i]) for i in range(n_dims)
                ]
                output_shape = [
                    utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                    for s in output_shape
                ]
                cropping_shape = [
                    int(np.around(output_shape[i] / resample_factor[i], 0))
                    for i in range(n_dims)
                ]
            # if no resampling, simply check if image_shape is divisible by n
            else:
                cropping_shape = [
                    utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                    for s in labels_shape
                ]
                output_shape = cropping_shape

        # if no need to be divisible by n, simply take cropping_shape as image_shape, and build output_shape
        else:
            cropping_shape = labels_shape
            if resample_factor is not None:
                output_shape = [
                    int(cropping_shape[i] * resample_factor[i]) for i in range(n_dims)
                ]
            else:
                output_shape = cropping_shape

    return cropping_shape, output_shape



"""
coregister - Image Registration for Aligning Neuroimaging Data

Part of the micaflow processing pipeline for neuroimaging data.

This module performs comprehensive image registration between two images using the
Advanced Normalization Tools (ANTs) SyNRA algorithm, which combines rigid, affine,
and symmetric normalization transformations. It aligns a moving image with a fixed
reference space, enabling spatial normalization of neuroimaging data for group analysis,
multimodal integration, or atlas-based analyses.

Features:
--------
- Combined rigid, affine, and SyN nonlinear registration in one step
- Bidirectional transformation capability (forward and inverse)
- Option to save all transformation components for later application
- Uses ANTs' powerful SyNRA algorithm for optimal accuracy
- Preserves header information in the registered output images

API Usage:
---------
micaflow coregister
    --fixed-file <path/to/reference.nii.gz>
    --moving-file <path/to/source.nii.gz>
    --output <path/to/registered.nii.gz>
    [--warp-file <path/to/warp.nii.gz>]
    [--affine-file <path/to/affine.mat>]
    [--rev-warp-file <path/to/reverse_warp.nii.gz>]
    [--rev-affine-file <path/to/reverse_affine.mat>]

Python Usage:
-----------
>>> from micaflow.scripts.coregister import ants_linear_nonlinear_registration
>>> ants_linear_nonlinear_registration(
...     fixed_file="mni152.nii.gz",
...     moving_file="subject_t1w.nii.gz",
...     out_file="registered_t1w.nii.gz",
...     warp_file="warp.nii.gz",
...     affine_file="affine.mat",
...     rev_warp_file="reverse_warp.nii.gz",
...     rev_affine_file="reverse_affine.mat"
... )

"""

import ants
import argparse
import shutil
import sys
from colorama import init, Fore, Style
import os
import multiprocessing
import nibabel as nib
import numpy as np
import tempfile


init()

# Get number of available CPU cores
DEFAULT_THREADS = multiprocessing.cpu_count()


def print_help():
    """Print a help message with examples."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                      IMAGE COREGISTRATION                      ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}

    This script performs linear (rigid + affine) and nonlinear (SyN) registration 
    between two images using ANTs. The registration aligns the moving image to 
    match the fixed reference image space.

    {CYAN}{BOLD}────────────────────────── USAGE ──────────────────────────{RESET}
      micaflow coregister {GREEN}[options]{RESET}

    {CYAN}{BOLD}─────────────────── REQUIRED ARGUMENTS ───────────────────{RESET}
      {YELLOW}--fixed{RESET}   : Path to the fixed/reference image (.nii.gz)
      {YELLOW}--moving{RESET}  : Path to the moving image to be registered (.nii.gz)

    {CYAN}{BOLD}─────────────────── OPTIONAL ARGUMENTS ───────────────────{RESET}
      {YELLOW}--output{RESET}             : Output path for the registered image (.nii.gz)
      {YELLOW}--registration-method{RESET}: Registration method (default: "SyNRA")
      {YELLOW}--warp-file{RESET}          : Path to save the forward warp field (.nii.gz)
      {YELLOW}--affine-file{RESET}        : Path to save the forward affine transform (.mat)
      {YELLOW}--inverse-warp-file{RESET}  : Path to save the reverse warp field (.nii.gz)
      {YELLOW}--inverse-affine-file{RESET}: Path to save the reverse affine transform (.mat)
      {YELLOW}--initial-affine-file{RESET}: Path to initial affine transform to use (.mat)
      {YELLOW}--initial-warp-file{RESET}  : Path to initial warp field to use (.nii.gz)
      {YELLOW}--interpolator{RESET}       : Interpolation method (default: "genericLabel")
      {YELLOW}--threads{RESET}            : Number of CPU threads (default: all available)

    {CYAN}{BOLD}───────────────── ANTS REGISTRATION OPTIONS ────────────────{RESET}
      {YELLOW}--verbose{RESET}             : Enable verbose output
      {YELLOW}--grad-step{RESET}           : Gradient step size (default: 0.2)
      {YELLOW}--flow-sigma{RESET}          : Smoothing for update field (default: 3)
      {YELLOW}--total-sigma{RESET}         : Smoothing for total field (default: 0)
      {YELLOW}--aff-metric{RESET}          : Metric for affine stage (default: "mattes")
      {YELLOW}--aff-sampling{RESET}        : Sampling parameter for affine metric (default: 32)
      {YELLOW}--syn-metric{RESET}          : Metric for SyN stage (default: "mattes")
      {YELLOW}--syn-sampling{RESET}        : Sampling parameter for SyN metric (default: 32)
      {YELLOW}--reg-iterations{RESET}      : SyN iterations, comma-separated (e.g., "40,20,0")
      {YELLOW}--aff-iterations{RESET}      : Affine iterations, comma-separated (e.g., "2100,1200,1200,10")
      {YELLOW}--aff-shrink-factors{RESET}  : Affine shrink factors, comma-separated (e.g., "6,4,2,1")
      {YELLOW}--aff-smoothing-sigmas{RESET}: Affine smoothing sigmas, comma-separated (e.g., "3,2,1,0")
      {YELLOW}--random-seed{RESET}         : Random seed for reproducibility

    {CYAN}{BOLD}────────────────── EXAMPLE USAGE ────────────────────────{RESET}

    {BLUE}# Basic registration with default parameters{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz

    {BLUE}# Registration with saved transforms{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz {YELLOW}--warp-file{RESET} warp.nii.gz {YELLOW}--affine-file{RESET} affine.mat

    {BLUE}# Registration with custom ANTs parameters{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz {YELLOW}--reg-iterations{RESET} "100,50,20" \\
      {YELLOW}--grad-step{RESET} 0.1 {YELLOW}--verbose{RESET}

    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────{RESET}
    {MAGENTA}•{RESET} The registration performs SyNRA transformation (rigid+affine+SyN)
    {MAGENTA}•{RESET} Forward transforms convert from moving space to fixed space
    {MAGENTA}•{RESET} Reverse transforms convert from fixed space to moving space
    {MAGENTA}•{RESET} The transforms can be applied to other images using apply_warp
    {MAGENTA}•{RESET} For reproducible results, set the random seed
    """
    print(help_text)


def combine_warps_and_transform(
    first_warp_path, 
    second_warp_path,
    combined_warp_output_path
):
    """
    Combines two warp fields, resampling the first warp to match the second warp if needed.
    
    Args:
        first_warp_path (str): Path to the first warp field
        second_warp_path (str): Path to the second warp field
        combined_warp_output_path (str): Path to save the combined warp field
    """
    # Load warp fields
    first_warp_nib = nib.load(first_warp_path)
    second_warp_nib = nib.load(second_warp_path)
    # Check if dimensions or affine transformations differ
    dims_match = (first_warp_nib.shape == second_warp_nib.shape)
    affine_match = np.allclose(first_warp_nib.affine, second_warp_nib.affine)
    
    if not (dims_match and affine_match):
        print(f"Resampling first warp to match second warp's dimensions and affine...")
        
        # Extract each component (x, y, z) and resample separately
        components = []
        for i in range(second_warp_nib.shape[-1]):  # Use the second warp's shape
            # Create temporary files for each component using tempfile module
            with tempfile.NamedTemporaryFile(suffix=f"_first_comp_{i}.nii.gz", delete=False) as tmp_first:
                first_comp_file = tmp_first.name
            
            with tempfile.NamedTemporaryFile(suffix=f"_second_comp_{i}.nii.gz", delete=False) as tmp_second:
                second_comp_file = tmp_second.name
            
            # Save component slices to temporary files
            first_comp_data = first_warp_nib.get_fdata()[..., i]
            first_comp_nib = nib.Nifti1Image(first_comp_data, first_warp_nib.affine)
            first_comp_nib.to_filename(first_comp_file)
            
            second_comp_data = second_warp_nib.get_fdata()[..., i]
            second_comp_nib = nib.Nifti1Image(second_comp_data, second_warp_nib.affine)
            second_comp_nib.to_filename(second_comp_file)
            
            # Read the components using ANTs
            first_comp = ants.image_read(first_comp_file)
            second_comp = ants.image_read(second_comp_file)
            
            # Resample first component to match second
            resampled_comp = ants.resample_image_to_target(
                first_comp, 
                second_comp,
                interp_type='linear'
            )
            
            # Store the resampled component
            components.append(resampled_comp.numpy())
            
            # Clean up temporary files
            try:
                os.remove(first_comp_file)
                os.remove(second_comp_file)
            except:
                pass
        
        # Combine components into a single array
        resampled_first_arr = np.stack(components, axis=-1)
        first_arr = resampled_first_arr
        second_arr = second_warp_nib.get_fdata().squeeze()
    else:
        # If dimensions match, just get the data
        first_arr = first_warp_nib.get_fdata().squeeze()
        second_arr = second_warp_nib.get_fdata().squeeze()

    # Add the displacements (this combines the transforms)
    combined_arr = first_arr + second_arr
    
    # Create and save the combined warp field
    combined_warp = nib.Nifti1Image(combined_arr, second_warp_nib.affine, second_warp_nib.header)
    combined_warp.to_filename(combined_warp_output_path)
    print(f"Combined warp field saved as {combined_warp_output_path}")


def ants_linear_nonlinear_registration(
    fixed_file,
    moving_file,
    out_file=None,
    warp_file=None,
    affine_file=None,
    rev_warp_file=None,
    rev_affine_file=None,
    registration_method="SyNRA",
    initial_affine_file=None,
    initial_warp_file=None,
    interpolator="genericLabel",
    threads=DEFAULT_THREADS,  # Use all available cores by default
    **kwargs,  # Add this to accept extra parameters
):
    """Perform linear (rigid + affine) and nonlinear registration using ANTsPy.

    This function performs registration between two images using ANTs' SyNRA transform,
    which includes both linear (rigid + affine) and nonlinear (SyN) components.

    Args:
        fixed_file (str): Path to the fixed/reference image.
        moving_file (str): Path to the moving image that will be registered.
        out_file (str, optional): Path where the registered image will be saved.
        warp_file (str, optional): Path to save the forward warp field.
        affine_file (str, optional): Path to save the forward affine transform.
        rev_warp_file (str, optional): Path to save the reverse warp field.
        rev_affine_file (str, optional): Path to save the reverse affine transform.
        registration_method (str): Registration method to use. Defaults to "SyNRA".
        initial_affine_file (str, optional): Path to initial affine transform.
        initial_warp_file (str, optional): Path to initial warp field.
        interpolator (str): Interpolation method. Defaults to "genericLabel".
        threads (int): Number of threads to use for registration. Defaults to all available cores.
        **kwargs: Additional arguments passed directly to ants.registration
                 Examples: verbose, grad_step, reg_iterations, etc.

    Returns:
        None: The function saves the registered image and transform files to disk.
    """
    if (
        not out_file
        and not warp_file
        and not affine_file
        and not rev_warp_file
        and not rev_affine_file
    ):
        print(Fore.RED + "Error: No outputs specified." + Style.RESET_ALL)
        sys.exit(1)

    # Set ANTs/ITK thread count in environment variables
    os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(threads)
    os.environ["OMP_NUM_THREADS"] = str(threads)

    # Load images
    fixed = ants.image_read(fixed_file)
    moving = ants.image_read(moving_file)
    initial_transform = []

    if initial_warp_file:
        initial_transform.append(initial_warp_file)

    if initial_affine_file:
        initial_transform.append(initial_affine_file)

    if initial_transform == []:
        initial_transform = None


    # Pass all arguments to ants.registration, including any extra kwargs
    transforms = ants.registration(
        fixed=fixed,
        moving=moving,
        type_of_transform=registration_method,
        # interpolator=interpolator,
        initial_transform=initial_transform,
        **kwargs,  # Pass through all additional arguments
    )
    print("Applying transforms...")
    registered = ants.apply_transforms(
        fixed=fixed,
        moving=moving,
        transformlist=transforms["fwdtransforms"],
        interpolator=interpolator,
    )
    print("Transforms applied successfully.")
    # Save the registered moving image
    if out_file is not None:
        ants.image_write(registered, out_file)
        print(f"Registration complete. Saved registered image as {out_file}")

    if warp_file:
        if initial_warp_file:
            combine_warps_and_transform(
                transforms["fwdtransforms"][0],
                transforms["fwdtransforms"][1],
                warp_file,)
            print(f"Saved warp field as {warp_file}")
        else:
            shutil.copyfile(transforms["fwdtransforms"][0], warp_file)
            print(f"Saved warp field as {warp_file}")

    if affine_file:
        if initial_affine_file:
            shutil.copyfile(transforms["fwdtransforms"][2], affine_file)
            print(f"Saved affine transform as {affine_file}")
        else:
            shutil.copyfile(transforms["fwdtransforms"][1], affine_file)
            print(f"Saved affine transform as {affine_file}")
    if rev_warp_file:
        if initial_warp_file:
            combine_warps_and_transform(
                transforms["invtransforms"][2],
                transforms["invtransforms"][1],
                warp_file,)
        else:
            shutil.copyfile(transforms["invtransforms"][1], rev_warp_file)
            print(f"Saved reverse warp field as {rev_warp_file}")
    if rev_affine_file:
        shutil.copyfile(transforms["invtransforms"][0], rev_affine_file)
        print(f"Saved reverse affine transform as {rev_affine_file}")

    print("All specified outputs saved successfully.")
    print("Cleaning up temporary files...")
    temp_files_to_delete = set(transforms['fwdtransforms'] + transforms['invtransforms'])
    deleted_count = 0
    for temp_file in temp_files_to_delete:
        try:
            if os.path.exists(temp_file):
                os.remove(temp_file)
                deleted_count += 1
        except OSError as e:
            print(f"Warning: Could not remove temporary file {temp_file}: {e}")
    print(f"Successfully cleaned up {deleted_count} temporary files.")


def main():
    """Entry point for command-line use"""
    parser = argparse.ArgumentParser(description="Coregistration tool")

    # Required/standard arguments
    parser.add_argument("--fixed", required=True, help="Fixed/reference image file path")
    parser.add_argument("--moving", required=True, help="Moving image file path")
    parser.add_argument("--output", help="Output image file path")
    parser.add_argument(
        "--registration-method", default="SyNRA", help="Registration method"
    )
    parser.add_argument("--affine-file", help="Affine transformation file path")
    parser.add_argument("--warp-file", help="Warp field file path")
    parser.add_argument(
        "--inverse-warp-file", help="Inverse warp field file path"  # Standardized name
    )
    parser.add_argument(
        "--inverse-affine-file", help="Inverse affine transformation file path"  # Standardized name
    )
    parser.add_argument(
        "--initial-affine-file", help="Initial affine transformation file path"
    )
    parser.add_argument("--initial-warp-file", help="Initial warp field file path")
    parser.add_argument(
        "--interpolator", help="Interpolator type", default="genericLabel"
    )
    
    # Add threads parameter with default value of all cores
    parser.add_argument(
        "--threads", 
        type=int, 
        default=DEFAULT_THREADS, 
        help=f"Number of threads to use (default: {DEFAULT_THREADS} - all cores)"
    )

    # Add common ANTs registration parameters
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument(
        "--grad-step", type=float, default=0.2, help="Gradient step size (default: 0.2)"
    )
    parser.add_argument(
        "--flow-sigma",
        type=float,
        default=3,
        help="Smoothing for update field (default: 3)",
    )
    parser.add_argument(
        "--total-sigma",
        type=float,
        default=0,
        help="Smoothing for total field (default: 0)",
    )
    parser.add_argument(
        "--aff-metric",
        default="mattes",
        help="Metric for affine stage (default: mattes)",
    )
    parser.add_argument(
        "--aff-sampling",
        type=int,
        default=32,
        help="Sampling parameter for affine metric (default: 32)",
    )
    parser.add_argument(
        "--syn-metric", default="mattes", help="Metric for SyN stage (default: mattes)"
    )
    parser.add_argument(
        "--syn-sampling",
        type=int,
        default=32,
        help="Sampling parameter for SyN metric (default: 32)",
    )

    # More complex parameters that need special handling
    parser.add_argument(
        "--reg-iterations", help="SyN iterations, comma-separated (e.g., '40,20,0')"
    )
    parser.add_argument(
        "--aff-iterations",
        help="Affine iterations, comma-separated (e.g., '2100,1200,1200,10')",
    )
    parser.add_argument(
        "--aff-shrink-factors",
        help="Affine shrink factors, comma-separated (e.g., '6,4,2,1')",
    )
    parser.add_argument(
        "--aff-smoothing-sigmas",
        help="Affine smoothing sigmas, comma-separated (e.g., '3,2,1,0')",
    )
    parser.add_argument(
        "--random-seed", type=int, help="Random seed for reproducibility"
    )

    args = parser.parse_args()

    # Process tuple arguments from strings
    kwargs = {}

    if args.verbose:
        kwargs["verbose"] = True

    # Add standard numeric parameters
    for param in [
        "grad_step",
        "flow_sigma",
        "total_sigma",
        "aff_metric",
        "aff_sampling",
        "syn_metric",
        "syn_sampling",
    ]:
        param_value = getattr(args, param.replace("-", "_"))
        if param_value is not None:
            kwargs[param] = param_value

    # Convert comma-separated strings to tuples for complex parameters
    for param in [
        "reg_iterations",
        "aff_iterations",
        "aff_shrink_factors",
        "aff_smoothing_sigmas",
    ]:
        param_value = getattr(args, param.replace("-", "_"))
        if param_value:
            try:
                # Convert string "40,20,0" to tuple (40, 20, 0)
                kwargs[param] = tuple(int(x) for x in param_value.split(","))
            except ValueError:
                print(f"Error parsing {param}. Use comma-separated integers.")
                sys.exit(1)

    # Add random seed if specified
    if args.random_seed is not None:
        kwargs["random_seed"] = args.random_seed

    # Call the coregister function with all arguments
    ants_linear_nonlinear_registration(
        fixed_file=args.fixed,
        moving_file=args.moving,
        out_file=args.output,
        registration_method=args.registration_method,
        affine_file=args.affine_file,
        warp_file=args.warp_file,
        rev_warp_file=args.inverse_warp_file,  # Use standardized name 
        rev_affine_file=args.inverse_affine_file,  # Use standardized name
        initial_affine_file=args.initial_affine_file,
        initial_warp_file=args.initial_warp_file,
        interpolator=args.interpolator,
        threads=args.threads,  # Pass threads parameter
        **kwargs,  # Pass all the extra parameters
    )


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.read().splitlines()

setup(
    name="lamar",
    version="1.0.0",
    author="Ian Goodall-Halliwell, Paul Bautin, Nya Yazdi, Kevin Du, Raul R. Cruces",
    author_email="gooodallhalliwell@gmail.com",
    description="Label Augmented Modality Agnostic Registration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MICA-MNI/LAMAReg",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: CC-BY-NC License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=requirements,
    entry_points={
        "console_scripts": ["lamar=lamareg.cli:main"],
    },
    include_package_data=True,
)
